#!/usr/bin/env python3
import os
import sys
import cmd
import json
import logging
from typing import Optional

from keys import generate_static_key, load_static_key, save_contact, list_contacts, CONTACTS_FILE
from transport import Peer
from noise_handshake import NoiseXK
from messaging import Messenger

# Configure default logging
logger = logging.getLogger("noise_xk_cli")
logger.setLevel(logging.INFO)
handler = logging.StreamHandler()
formatter = logging.Formatter("[%(levelname)s] %(message)s")
handler.setFormatter(formatter)
logger.addHandler(handler)


class NoiseCLI(cmd.Cmd):
    intro = "Welcome to the Noise_XK CLI. Type help or ? to list commands."
    prompt = "(noise) "

    def __init__(self):
        super().__init__()
        # Static key pair for this instance
        self.static_priv = None
        self.static_pub = None
        # Peer static key (bytes)
        self.peer_static: Optional[bytes] = None
        # Transport peer connection
        self.peer: Optional[Peer] = None
        # Handshake and messenger
        self.handshake: Optional[NoiseXK] = None
        self.messenger: Optional[Messenger] = None
        # Incoming ciphertext buffer
        self.inbox = []
        # Ensure contacts directory
        os.makedirs(os.path.dirname(CONTACTS_FILE), exist_ok=True)

    def preloop(self):
        # Load existing static key or generate new
        priv, pub = generate_static_key()
        self.static_priv, self.static_pub = priv, pub
        logger.info(f"Generated static X25519 key, public: {pub.hex()}")

    def do_generate_key(self, arg):
        "Regenerate a new static key pair"
        priv, pub = generate_static_key()
        self.static_priv, self.static_pub = priv, pub
        logger.info(f"New static public key: {pub.hex()}")

    def do_import_key(self, arg):
        "import_key <alias> <hex_pub>  Import peer static public key"
        parts = arg.split()
        if len(parts) != 2:
            logger.error("Usage: import_key <alias> <hex_pub>")
            return
        alias, hex_pub = parts
        try:
            pub = load_static_key(hex_pub)
            save_contact(alias, hex_pub)
            logger.info(f"Imported contact '{alias}' = {hex_pub}")
        except Exception as e:
            logger.error(f"Failed to import key: {e}")

    def do_list_keys(self, arg):
        "List saved contacts"
        contacts = list_contacts()
        if not contacts:
            logger.info("No contacts saved.")
            return
        for alias, hex_pub in contacts:
            print(f"  {alias}: {hex_pub}")

    def do_listen(self, arg):
        "listen <port>  Start listening for incoming P2P connections"
        if not arg:
            logger.error("Usage: listen <port>")
            return
        try:
            port = int(arg.strip())
        except ValueError:
            logger.error("Port must be an integer.")
            return
        self.peer = Peer(port, self.handle_data)
        self.peer.start_listen()
        logger.info(f"Listening on port {port}")

    def do_connect(self, arg):
        "connect [host] <port> <alias>  Connect to remote peer and set peer static key by alias"
        parts = arg.split()
        if len(parts) == 2:
            # Host not specified, use localhost as default
            host = "localhost"
            port_str, alias = parts
        elif len(parts) == 3:
            # Host explicitly specified
            logger.debug(parts)
            host, port_str, alias = parts
        else:
            logger.error("Usage: connect [host] <port> <alias>")
            return
        try:
            port = int(port_str)
        except ValueError:
            logger.error("Port must be an integer. " + port_str)
            return

        contacts = dict(list_contacts())
        if alias not in contacts.keys():
            logger.error(f"No such alias '{alias}'. Use import_key first.")
            return

        self.peer_static = load_static_key(contacts[alias])
        # Establish connection
        self.peer = Peer(port, self.handle_data)
        try:
            self.peer.connect(host, port)
            logger.info(f"Connected to {host}:{port}")
        except Exception as e:
            logger.error(f"Connection failed: {e}")

    def do_handshake(self, arg):
        "handshake  Initiate XK handshake with the peer (initiator role)"
        if not self.peer or not self.peer_static:
            logger.error("Need to connect first and import peer key.")
            return
        # Create handshake as initiator
        self.handshake = NoiseXK(initiator=True, peer_static=self.peer_static)
        # Send first handshake message
        msg = self.handshake.write()
        self.peer.send(msg)
        logger.info("Sent handshake message 1")

    def handle_data(self, data: bytes):
        "Callback for incoming raw data"
        if self.handshake and not self.handshake.is_finished():
            # In the middle of handshake
            try:
                _ = self.handshake.read(data)
                if not self.handshake.is_finished():
                    # Send next handshake message
                    msg = self.handshake.write()
                    self.peer.send(msg)
                    logger.info("Exchanged handshake message")
                else:
                    cs_s, cs_r = self.handshake.get_cipher_states()
                    self.messenger = Messenger(cs_s, cs_r)
                    logger.info("Handshake complete, cipher states ready")
                return
            except Exception as e:
                logger.error(f"Handshake error: {e}")
                return

        # If no handshake object, data might be initial handshake for responder
        if not self.handshake and self.peer_static:
            # Passive responder: start handshake on first data
            try:
                self.handshake = NoiseXK(initiator=False, peer_static=self.peer_static)
                _ = self.handshake.read(data)
                # Send response
                msg = self.handshake.write()
                self.peer.send(msg)
                # Continue handshake if needed
                if not self.handshake.is_finished():
                    # Await next, then finalize
                    logger.info("Responder sent handshake message 2")
                else:
                    cs_s, cs_r = self.handshake.get_cipher_states()
                    self.messenger = Messenger(cs_s, cs_r)
                    logger.info("Handshake complete (responder), cipher states ready")
                return
            except Exception as e:
                logger.error(f"Responder handshake error: {e}")
                return

        # After handshake, treat as encrypted application data
        if self.messenger:
            try:
                plaintext = self.messenger.decrypt(data)
                logger.info(f"Received plaintext: {plaintext.decode(errors='ignore')}")
            except Exception as e:
                # If decryption fails, queue raw ciphertext
                self.inbox.append(data)
                idx = len(self.inbox) - 1
                logger.warning(f"Failed decrypt, queued ciphertext at index {idx}")
        else:
            logger.warning("Received data but handshake not established.")

    def do_send(self, arg):
        "send <text>  Encrypt and send a message to the peer"
        if not self.messenger:
            logger.error("Handshake not complete.")
            return
        payload = arg.encode()
        ct = self.messenger.encrypt(payload)
        self.peer.send(ct)
        logger.info(f"Sent encrypted message ({len(ct)} bytes)")

    def do_decrypt(self, arg):
        "decrypt <index>  Attempt to decrypt a queued ciphertext by index"
        if not self.messenger:
            logger.error("Handshake not complete.")
            return
        try:
            idx = int(arg.strip())
            ct = self.inbox[idx]
            pt = self.messenger.decrypt(ct)
            logger.info(f"Decrypted [#{idx}]: {pt.decode(errors='ignore')}")
        except Exception as e:
            logger.error(f"Decrypt error: {e}")

    def do_show_state(self, arg):
        "show_state  Display current internal state"
        state = {
            "static_pub": self.static_pub.hex() if self.static_pub else None,
            "peer_static": self.peer_static.hex() if self.peer_static else None,
            "handshake": type(self.handshake).__name__ if self.handshake else None,
            "handshake_finished": self.handshake.is_finished() if self.handshake else False,
            "inbox_size": len(self.inbox),
        }
        print(json.dumps(state, indent=2))

    def do_set_log_level(self, arg):
        "set_log_level <DEBUG|INFO|WARNING|ERROR>  Adjust CLI verbosity"
        level = arg.strip().upper()
        if level not in ("DEBUG","INFO","WARNING","ERROR"):
            logger.error("Invalid level. Choose DEBUG, INFO, WARNING, ERROR.")
            return
        logger.setLevel(getattr(logging, level))
        logger.info(f"Log level set to {level}")

    def do_exit(self, arg):
        "Exit the CLI"
        logger.info("Closing connection and exiting.")
        if self.peer:
            self.peer.close()
        return True

    def do_EOF(self, arg):
        return self.do_exit(arg)


if __name__ == "__main__":
    # Adjust module search path
    sys.path.insert(0, os.path.abspath(os.path.dirname(__file__)))
    from transport import Peer  # ensure imports
    cli = NoiseCLI()
    cli.cmdloop()
