"""
Module keys.py

Génération, import, export et gestion de clés statiques X25519 pour Noise_XK CLI.
Stocke les contacts dans ~/.noise_xk_cli/contacts.json
"""
import os
import json
from typing import Tuple, Dict, List

from cryptography.hazmat.primitives.asymmetric.x25519 import (
    X25519PrivateKey,
    X25519PublicKey,
)
from cryptography.hazmat.primitives import serialization

# Chemin vers le fichier de contacts
CONTACTS_FILE = os.path.expanduser("~/.noise_xk_cli/contacts.json")


def _ensure_contacts_file() -> None:
    """
    Crée le répertoire et le fichier JSON si nécessaire.
    """
    directory = os.path.dirname(CONTACTS_FILE)
    os.makedirs(directory, exist_ok=True)
    if not os.path.exists(CONTACTS_FILE):
        with open(CONTACTS_FILE, 'w') as f:
            json.dump({}, f)


def generate_static_key() -> Tuple[X25519PrivateKey, bytes]:
    """
    Génère une paire de clés X25519 statique.

    Returns:
        (private_key, public_key_bytes)
    """
    private_key = X25519PrivateKey.generate()
    public_key = private_key.public_key()
    public_bytes = public_key.public_bytes(
        encoding=serialization.Encoding.Raw,
        format=serialization.PublicFormat.Raw
    )
    return private_key, public_bytes


def public_key_to_hex(public_bytes: bytes) -> str:
    """
    Convertit bytes de clé publique en hex.
    """
    return public_bytes.hex()


def load_static_key(hex_str: str) -> X25519PublicKey:
    """
    Charge une clé publique X25519 depuis sa représentation hex.
    """
    data = bytes.fromhex(hex_str)
    if len(data) != 32:
        raise ValueError("Clé publique X25519 doit faire 32 octets (raw)")
    return X25519PublicKey.from_public_bytes(data)


def save_contact(alias: str, pub_hex: str) -> None:
    """
    Sauvegarde un contact dans le fichier JSON sous l'alias donné.
    """
    _ensure_contacts_file()
    with open(CONTACTS_FILE, 'r+') as f:
        data: Dict[str, str] = json.load(f)
        data[alias] = pub_hex
        f.seek(0)
        json.dump(data, f, indent=2)
        f.truncate()


def list_contacts() -> List[Tuple[str, str]]:
    """
    Retourne la liste des contacts (alias, clé publique hex).
    """
    _ensure_contacts_file()
    with open(CONTACTS_FILE, 'r') as f:
        data: Dict[str, str] = json.load(f)
    return list(data.items())


def get_contact(alias: str) -> str:
    """
    Récupère la clé publique hex pour un alias donné.

    Raises:
        KeyError: si l'alias n'existe pas
    """
    _ensure_contacts_file()
    with open(CONTACTS_FILE, 'r') as f:
        data: Dict[str, str] = json.load(f)
    if alias not in data:
        raise KeyError(f"Alias '{alias}' introuvable dans les contacts")
    return data[alias]
