# noise_handshake.py

from dissononce.extras.meta.protocol.factory import NoiseProtocolFactory
# from dissononce.exceptions import HandshakeError

class NoiseXK:
    """
    Wrapper around the Noise_XK_25519_ChaChaPoly_SHA256 handshake pattern,
    using dissononce rev.34.
    """

    def __init__(self, initiator: bool, peer_static: bytes):
        """
        :param initiator: True if this party initiates the handshake.
        :param peer_static: The peer's static public key (32 bytes X25519).
        """
        # Instantiate the factory for XK over 25519/ChaChaPoly/SHA256
        factory = NoiseProtocolFactory.from_name(
            "Noise_XK_25519_ChaChaPoly_SHA256"
        )

        # Create the HandshakeState:
        # - if initiator: supply 's' = peer_static
        # - if responder: supply 'rs' = peer_static
        self._hs = factory.create_handshakestate(
            initiator=initiator,
            s=peer_static if initiator else None,
            rs=None if initiator else peer_static
        )

    def write(self, payload: bytes = b"") -> bytes:
        """
        Generate the next handshake message to send.
        :param payload: optional application data to embed (usually b"").
        :return: the raw bytes of the handshake message.
        """
        try:
            msg = self._hs.write_message(payload)
            return msg
        except Exception as e:
            raise Exception(f"Error writing handshake message: {e}") from e

    def read(self, message: bytes) -> bytes:
        """
        Process a received handshake message.
        :param message: raw bytes received from the peer.
        :return: any payload decrypted from the message (usually b"").
        """
        try:
            payload = self._hs.read_message(message)
            return payload
        except Exception as e:
            raise Exception(f"Error reading handshake message: {e}") from e

    def is_finished(self) -> bool:
        """
        :return: True if the handshake is complete and cipher-states are ready.
        """
        return self._hs.is_finished()

    def get_cipher_states(self):
        """
        Split the HandshakeState into two CipherState objects once the handshake is done.
        :return: (cs_send, cs_recv)
        :raises: HandshakeError if called before handshake completion.
        """
        if not self._hs.is_finished():
            raise Exception("Cannot split cipher-states before handshake is finished")
        return self._hs.split()
