"""
Module transport.py

Abstraction pair-à-pair TCP basique pour CLI Noise_XK.
Chaque instance peut écouter sur un port et/ou se connecter à un pair.
"""
import socket
import threading
from typing import Callable, Optional


class Peer:
    """
    Peer TCP abstrait : écoute et connexion à un seul pair.
    on_data(data: bytes) est appelé à la réception de données.
    """
    def __init__(self, port: int, on_data: Callable[[bytes], None], listen: bool = True):
        self.port = port
        self.on_data = on_data
        self.server_socket: Optional[socket.socket] = None
        self.conn_socket: Optional[socket.socket] = None
        self.alive = False
        if listen:
            self.start_listen()

    def start_listen(self) -> None:
        """
        Démarre le socket serveur et la boucle d'accept.
        """
        if self.server_socket:
            return
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
        sock.bind(('0.0.0.0', self.port))
        sock.listen(1)
        self.server_socket = sock
        self.alive = True
        threading.Thread(target=self._accept_loop, daemon=True).start()

    def _accept_loop(self) -> None:
        while self.alive:
            try:
                client_sock, addr = self.server_socket.accept()
                # Fermer connexion précédente si existante
                if self.conn_socket:
                    self.conn_socket.close()
                self.conn_socket = client_sock
                threading.Thread(
                    target=self._read_loop,
                    args=(client_sock,),
                    daemon=True
                ).start()
            except OSError:
                break

    def connect(self, host: str, port: int) -> None:
        """
        Se connecte à un peer à l'adresse spécifiée.
        """
        sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        sock.connect((host, port))
        # Fermer connexion précédente si existante
        if self.conn_socket:
            self.conn_socket.close()
        self.conn_socket = sock
        self.alive = True
        threading.Thread(
            target=self._read_loop,
            args=(sock,),
            daemon=True
        ).start()

    def _read_loop(self, sock: socket.socket) -> None:
        """
        Boucle de lecture asynchrone sur la socket.
        """
        while self.alive:
            try:
                data = sock.recv(4096)
                if not data:
                    break
                self.on_data(data)
            except OSError:
                break
        try:
            sock.close()
        except OSError:
            pass

    def send(self, data: bytes) -> None:
        """
        Envoie de données au peer (connexion active).
        """
        if not self.conn_socket:
            raise ConnectionError("Aucune connexion active pour envoyer")
        self.conn_socket.sendall(data)

    def close(self) -> None:
        """
        Ferme sockets d'écoute et de connexion.
        """
        self.alive = False
        if self.server_socket:
            try:
                self.server_socket.close()
            except OSError:
                pass
            self.server_socket = None
        if self.conn_socket:
            try:
                self.conn_socket.close()
            except OSError:
                pass
            self.conn_socket = None
